"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformThrowStatement = exports.transformTryStatement = void 0;
const __1 = require("../..");
const lua = require("../../LuaAST");
const diagnostics_1 = require("../utils/diagnostics");
const lua_ast_1 = require("../utils/lua-ast");
const scope_1 = require("../utils/scope");
const typescript_1 = require("../utils/typescript");
const block_1 = require("./block");
const identifier_1 = require("./identifier");
const multi_1 = require("./language-extensions/multi");
const return_1 = require("./return");
const transformTryStatement = (statement, context) => {
    const [tryBlock, tryScope] = (0, block_1.transformScopeBlock)(context, statement.tryBlock, scope_1.ScopeType.Try);
    if (context.options.luaTarget === __1.LuaTarget.Lua51 && (0, typescript_1.isInAsyncFunction)(statement)) {
        context.diagnostics.push((0, diagnostics_1.unsupportedForTarget)(statement, "try/catch inside async functions", __1.LuaTarget.Lua51));
        return tryBlock.statements;
    }
    if (context.options.luaTarget === __1.LuaTarget.Lua51 && (0, typescript_1.isInGeneratorFunction)(statement)) {
        context.diagnostics.push((0, diagnostics_1.unsupportedForTarget)(statement, "try/catch inside generator functions", __1.LuaTarget.Lua51));
        return tryBlock.statements;
    }
    const tryResultIdentifier = lua.createIdentifier("____try");
    const returnValueIdentifier = lua.createIdentifier("____returnValue");
    const result = [];
    const returnedIdentifier = lua.createIdentifier("____hasReturned");
    let returnCondition;
    const pCall = lua.createIdentifier("pcall");
    const tryCall = lua.createCallExpression(pCall, [lua.createFunctionExpression(tryBlock)]);
    if (statement.catchClause && statement.catchClause.block.statements.length > 0) {
        // try with catch
        const [catchBlock, catchScope] = (0, block_1.transformScopeBlock)(context, statement.catchClause.block, scope_1.ScopeType.Catch);
        const catchParameter = statement.catchClause.variableDeclaration
            ? (0, identifier_1.transformIdentifier)(context, statement.catchClause.variableDeclaration.name)
            : undefined;
        const catchParameters = () => (catchParameter ? [lua.cloneIdentifier(catchParameter)] : []);
        const catchIdentifier = lua.createIdentifier("____catch");
        const catchFunction = lua.createFunctionExpression(catchBlock, catchParameters());
        result.push(lua.createVariableDeclarationStatement(catchIdentifier, catchFunction));
        const tryReturnIdentifiers = [tryResultIdentifier]; // ____try
        if (returnedIdentifier) {
            tryReturnIdentifiers.push(returnedIdentifier); // ____returned or catch variable
            if (tryScope.functionReturned || catchScope.functionReturned) {
                tryReturnIdentifiers.push(returnValueIdentifier); // ____returnValue
                returnCondition = lua.cloneIdentifier(returnedIdentifier);
            }
        }
        result.push(lua.createVariableDeclarationStatement(tryReturnIdentifiers, tryCall));
        // Wrap catch in function if try or catch has return
        const catchCall = lua.createCallExpression(catchIdentifier, [lua.cloneIdentifier(returnedIdentifier)]);
        const catchAssign = lua.createAssignmentStatement([lua.cloneIdentifier(returnedIdentifier), lua.cloneIdentifier(returnValueIdentifier)], catchCall);
        const notTryCondition = lua.createUnaryExpression(tryResultIdentifier, lua.SyntaxKind.NotOperator);
        result.push(lua.createIfStatement(notTryCondition, lua.createBlock([catchAssign])));
    }
    else if (tryScope.functionReturned) {
        // try with return, but no catch
        // returnedIdentifier = lua.createIdentifier("____returned");
        const returnedVariables = [tryResultIdentifier, returnedIdentifier, returnValueIdentifier];
        result.push(lua.createVariableDeclarationStatement(returnedVariables, tryCall));
        // change return condition from '____returned' to '____try and ____returned'
        returnCondition = lua.createBinaryExpression(lua.cloneIdentifier(tryResultIdentifier), returnedIdentifier, lua.SyntaxKind.AndOperator);
    }
    else {
        // try without return or catch
        result.push(lua.createExpressionStatement(tryCall));
    }
    if (statement.finallyBlock && statement.finallyBlock.statements.length > 0) {
        result.push(...context.transformStatements(statement.finallyBlock));
    }
    if (returnCondition && returnedIdentifier) {
        const returnValues = [];
        if ((0, multi_1.isInMultiReturnFunction)(context, statement)) {
            returnValues.push((0, lua_ast_1.createUnpackCall)(context, lua.cloneIdentifier(returnValueIdentifier)));
        }
        else {
            returnValues.push(lua.cloneIdentifier(returnValueIdentifier));
        }
        const returnStatement = (0, return_1.createReturnStatement)(context, returnValues, statement);
        const ifReturnedStatement = lua.createIfStatement(returnCondition, lua.createBlock([returnStatement]));
        result.push(ifReturnedStatement);
    }
    return lua.createDoStatement(result, statement);
};
exports.transformTryStatement = transformTryStatement;
const transformThrowStatement = (statement, context) => {
    const parameters = [];
    if (statement.expression) {
        parameters.push(context.transformExpression(statement.expression));
        parameters.push(lua.createNumericLiteral(0));
    }
    return lua.createExpressionStatement(lua.createCallExpression(lua.createIdentifier("error"), parameters), statement);
};
exports.transformThrowStatement = transformThrowStatement;
//# sourceMappingURL=errors.js.map